// Tenta aumentar a variabilidade do cardápio fazendo troca de pratos entre os dias em uma mesma cooluna
const moment = require('moment');
const DataManagement = require('./DataManagement');
this.seqMap = null;
const dayOfWeek = {
  1: 'IND_SEGUNDA',
  2: 'IND_TERCA',
  3: 'IND_QUARTA',
  4: 'IND_QUINTA',
  5: 'IND_SEXTA',
  6: 'IND_SABADO',
  7: 'IND_DOMINGO',
};
analisaCardapio = (cardapio, seqsMap) => {
  let groupMap = new Map();
  let diasSeq = new Map();
  let gruposDuplicadosDia = new Map();
  cardapio.datas.forEach((dia) => {
    const TODAY = dia.dia;
    const YESTERDAY = moment(dia.dia, 'DD/MM/YYYY').subtract(1, 'days').format('DD/MM/YYYY');
    let groupsCount = {};
    let gruposDuplicados = [];
    const SEQ_GRUPOS = dia.pratos.reduce((acc, { sequencial, gpprato }) => {
      acc.set(gpprato, null);
      if (seqsMap.get(sequencial) > 1) {
        // possui mais grupos possiveis para o sequencial
        if (!groupsCount[gpprato]) {
          groupsCount[gpprato] = 1;
        } else {
          groupsCount[gpprato]++;
          gruposDuplicados.push(gpprato);
        }
      }
      return acc;
    }, new Map());
    if (gruposDuplicados.length) {
      gruposDuplicadosDia.set(TODAY, gruposDuplicados[0]); // trata por hora apenas primeiro grupo duplicado
    }
    diasSeq.set(TODAY, SEQ_GRUPOS);
    if (diasSeq.has(YESTERDAY)) {
      // verifica dias consecutivos
      let consecutivos = Array.from(diasSeq.get(TODAY).keys()).filter((todayGroupSeq) =>
        diasSeq.get(YESTERDAY).has(todayGroupSeq),
      );
      consecutivos.forEach((seqGroup) => {
        const getArrDays = (groupMap, seqGroup, yesterday) => {
          if (groupMap.has(seqGroup)) {
            let arr = groupMap.get(seqGroup).find((arrDays) => arrDays[arrDays.length - 1] == yesterday);
            if (!arr) {
              groupMap.get(seqGroup).push([]);
              return groupMap.get(seqGroup)[groupMap.get(seqGroup).length - 1];
            }
            return arr;
          } else {
            groupMap.set(seqGroup, [[]]);
            return groupMap.get(seqGroup)[0];
          }
        };
        let arrayToInsert = getArrDays(groupMap, seqGroup, YESTERDAY);
        arrayToInsert.push(TODAY);
        if (!arrayToInsert.find((day) => day == YESTERDAY)) arrayToInsert.unshift(YESTERDAY);
      });
    }
  });
  return { groupMap, gruposDuplicadosDia };
};

const priorizaDiasIntercalados = (arr) => {
  arr.sort((a, b) => {
    return arr.findIndex((el) => el == a) % 2 != 0 ? -1 : 1;
  });
  return arr;
};

const getDay = (date) => {
  return parseInt(date.substring(0, 2));
};

const retornaCardapioSemana = (cardapio, diaAtual, diasIgnorados, ignorarDiaAtual = false) => {
  let diaObj = moment(diaAtual, 'DD/MM/YYYY');
  let startWeek = diaObj.clone().startOf('week');
  let startWeekDay = startWeek.month() !== diaObj.month() ? 1 : parseInt(startWeek.format('DD'));
  let endWeek = diaObj.clone().endOf('week');
  let endWeekDay = endWeek.month() !== diaObj.month() ? diaObj.daysInMonth() : parseInt(endWeek.format('DD'));
  let indexCardDia = null;
  let cardapioFiltrado = cardapio.datas.filter(({ dia }, index) => {
    let today = parseInt(dia.substring(0, 2));
    let inWeek = (today > startWeekDay && today < endWeekDay) || today == startWeekDay || today == endWeekDay;
    if (dia == diaAtual) {
      indexCardDia = index;
    }
    return (
      inWeek &&
      (!ignorarDiaAtual || (ignorarDiaAtual && dia !== diaAtual)) &&
      diasIgnorados.every((diaIg) => diaIg !== dia)
    );
  });
  return { cardapio: cardapioFiltrado, indexCardDia };
};

const possuiDiasConsecutivos = (datas) => {
  return !!datas
    .map(getDay)
    .sort((a, b) => a - b)
    .find((dia, index, array) => {
      return dia + 1 == array[index + 1];
    });
};

const pratoPossuiRestricaoIngrediente = (prato, restricaoIng) => {
  if (!prato.INGR_PRINCIPAL_1 && !prato.INGR_PRINCIPAL_2) return false;
  return restricaoIng.some((res) => {
    let sequenciais = res.TIPOPRAT.match(/\d+/gi) || [];
    if (!sequenciais.includes(prato.COD_TIP_PRATO)) return false;
    return (
      (res.INGR_PRINCIPAL_1 == 'S' &&
        (res.NRINGREDIENTEPRINC == null || prato.INGR_PRINCIPAL_1 == res.NRINGREDIENTEPRINC)) ||
      (res.INGR_PRINCIPAL_2 == 'S' &&
        (res.NRINGREDIENTEPRINC == null || prato.INGR_PRINCIPAL_2 == res.NRINGREDIENTEPRINC))
    );
  });
};

const checkUtilizacao = (ut, day) => {
  day.diasemana = moment(day.DTDATA, 'DD/MM/YYYY HH:mm:ss').isoWeekday();
  let diaSitu = [day.IDPRIMSEM, day.IDSEMANA2, day.IDSEMANA3, day.IDSEMANA4, day.IDULTMSEM];
  let utilizacoes = [ut.IDPRIMSEM, ut.IDSEMANA2, ut.IDSEMANA3, ut.IDSEMANA4, ut.IDULTMSEM];
  let obrigatorio = false;
  let restrito = false;
  let okDSem = ut[dayOfWeek[day.diasemana]] != 'R';
  let okFer = (day.IDFERIADO == 'S' && ut.IDFERIADO != 'R') || day.IDFERIADO == 'N';

  if (day.IDULTMSEM == 'S' && diaSitu.filter((d) => d == 'S').length == 2) {
    let indexes = [];
    diaSitu.forEach((d, k) => {
      if (d == 'S') {
        indexes.push(k);
      }
    });

    let newUt = [utilizacoes[indexes[0]], utilizacoes[indexes[1]]];
    obrigatorio = newUt[1] == 'O' ? true : newUt[1] == 'R' ? false : newUt[0] == 'O';
    restrito = newUt[1] == 'R' ? true : newUt[1] == 'O' ? false : newUt[0] == 'R';
  } else {
    obrigatorio = diaSitu.find((d, key) => d == 'S' && ['P', 'O'].includes(utilizacoes[key]));
    restrito = diaSitu.find((d, key) => d == 'S' && utilizacoes[key] == 'R');
  }
  return {
    obrigatorio: obrigatorio === true || obrigatorio == 'S' || (day.IDFERIADO == 'S' && ut.IDFERIADO == 'O'),
    restrito: !okDSem || restrito === true || restrito == 'S' || !okFer,
  };
};

const retornaUtilizacaoGrupoSemana = (dia, prato, utilizacaoPratoSemana) => {
  let utilizacao = utilizacaoPratoSemana.find((ut) => ut.NUM_GRP_PRATO == prato.gpprato);
  if (!utilizacao) return { obrigatorio: false, restrito: false };
  return checkUtilizacao(utilizacao, dia);
};

const pratoPossuiRestObg = (dia, prato, utilizacaoPratoSemana) => {
  const { obrigatorio, restrito } = retornaUtilizacaoGrupoSemana(dia, prato, utilizacaoPratoSemana);
  return obrigatorio || restrito;
};

const pratoPossuiRestricaoCombinacao = (prato, resPratoCombinacao) => {
  return resPratoCombinacao.some((res) => {
    let sequenciais = res.SEQ.match(/\d+/gi) || [];
    let isMatchSeq = sequenciais.includes(prato.sequencial);
    return (
      prato.gpprato == res.NRGRUPOPRATOSERV ||
      prato.prato == res.COD_PRATO ||
      prato.gpprato == res.NRGRUPOPRATOSERV_COMB ||
      prato.prato == res.COD_PRATO_COMB ||
      (isMatchSeq && (!res.CDTIPOPRAT || res.CDTIPOPRAT == prato.COD_TIP_PRATO))
    );
  });
};
const copyObj = (obj) => {
  return JSON.parse(JSON.stringify(obj));
};

const registraLog = (log, dia1, dia2, prato1, prato2, desc = '') => {
  log[dia1] = log[dia1] || [];
  log[dia1].push({
    dia: dia2,
    Antes: `${prato1.gppratoNome} | ${prato1.prato} - ${prato1.nomePrato}`,
    Coluna: this.seqMap && this.seqMap.has(prato1.sequencial) ? this.seqMap.get(prato1.sequencial) : ' - ',
    Depois: `${prato2.gppratoNome} | ${prato2.prato} - ${prato2.nomePrato}`,
    Obs: desc,
  });
};

const getComb = (cardapio, pratoDia) => {
  return cardapio.filter((p) => {
    return (
      (p.pratoOrigin && p.pratoOrigin === pratoDia.prato) ||
      (p.prato && p.prato === pratoDia.pratoOrigin) ||
      (pratoDia.pratoOrigin && pratoDia.pratoOrigin === p.pratoOrigin)
    );
  });
};

const verificaGrupoDiasConsecutivos = (cardapioTabela, pratoOrigem, pratoCandidato, diaOrigem, diaAtual) => {
  let possuiGrupoAntesDepois = false;
  if (diaAtual > 1 && diaAtual !== cardapioTabela.length - 1) {
    let possuiGrupoAntes = cardapioTabela[diaAtual - 1].some(
      (p) => p.gpprato == pratoOrigem.gpprato && isOtherDishOrDay(pratoOrigem, p),
    ); //dia anterior
    let possuiGrupoDepois = cardapioTabela[diaAtual + 1].some(
      (p) => p.gpprato == pratoOrigem.gpprato && isOtherDishOrDay(pratoOrigem, p),
    ); // dia posterior
    if (possuiGrupoAntes || possuiGrupoDepois) possuiGrupoAntesDepois = true;
  } else if (diaAtual == 1) {
    // primeiroDia
    let possuiGrupoDepois = cardapioTabela[diaAtual + 1].some(
      (p) => p.gpprato == pratoOrigem.gpprato && isOtherDishOrDay(pratoOrigem, p),
    );
    if (possuiGrupoDepois) possuiGrupoAntesDepois = true;
  } else if (diaAtual == cardapioTabela.length - 1) {
    // ultimoDia
    let possuiGrupoAntes = cardapioTabela[diaAtual - 1].some(
      (p) => p.gpprato == pratoOrigem.gpprato && isOtherDishOrDay(pratoOrigem, p),
    );
    if (possuiGrupoAntes) possuiGrupoAntesDepois = true;
  }
  if (!possuiGrupoAntesDepois) {
    if (diaOrigem > 1 && diaOrigem !== cardapioTabela.length - 1) {
      let possuiGrupoAntes = cardapioTabela[diaOrigem - 1].some(
        (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
      ); //dia anterior
      let possuiGrupoDepois = cardapioTabela[diaOrigem + 1].some(
        (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
      ); // dia posterior
      if (possuiGrupoAntes || possuiGrupoDepois) possuiGrupoAntesDepois = true;
    } else if (diaOrigem == 1) {
      // primeiroDia
      let possuiGrupoDepois = cardapioTabela[diaOrigem + 1].some(
        (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
      ); // dia posterior
      if (possuiGrupoDepois) possuiGrupoAntesDepois = true;
    } else if (diaOrigem == cardapioTabela.length - 1) {
      // ultimoDia
      let possuiGrupoAntes = cardapioTabela[diaOrigem - 1].some(
        (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
      ); //dia anterior
      if (possuiGrupoAntes) possuiGrupoAntesDepois = true;
    }
  }
  return possuiGrupoAntesDepois;
};

const isOtherDishOrDay = (primaryDish, secondDish) => {
  return (
    primaryDish.data !== secondDish.data ||
    primaryDish.gpprato !== secondDish.gpprato ||
    primaryDish.prato !== secondDish.prato
  );
};

const realizaTrocaPrato = (
  cardapio,
  cardapioDia,
  cardapioTabela,
  cardapioSemana,
  pratoDia,
  originalData,
  groupSeqMap,
  log,
  force = false,
) => {
  let {
    res_prato_combinacao: resPratoCombinacao,
    restricao_de_repeticao_ingrediente: restricaoIngrediente,
    utilizacao_prato_semana: utilizacaoPratoSemana,
  } = originalData;
  let seqOrigem = pratoDia.sequencial;
  resPratoCombinacao = resPratoCombinacao.data;
  restricaoIngrediente = restricaoIngrediente.data;
  utilizacaoPratoSemana = utilizacaoPratoSemana.data;

  if (pratoPossuiRestricaoCombinacao(pratoDia, resPratoCombinacao)) return false;
  if (pratoPossuiRestricaoIngrediente(pratoDia, restricaoIngrediente)) return false;
  let diasCandidatos = cardapioSemana.filter((dia) => {
    let diaCalendarioOrigem = originalData.calendario.data.find((d) => d.DTDATA.split(' ')[0] == dia.dia); // dia origem
    let diaCalendarioDestino = originalData.calendario.data.find((d) => d.DTDATA.split(' ')[0] == pratoDia.data); // dia destino
    let possuiGrupoDia = false;
    let pratosPossiveis = dia.pratos.filter((p) => {
      if (p.gpprato == pratoDia.gpprato) possuiGrupoDia = true;
      return (
        p.prato /*&& p.sequencial == pratoDia.sequencial */ &&
        p.gpprato !== pratoDia.gpprato &&
        !pratoPossuiRestricaoCombinacao(p, resPratoCombinacao) &&
        !pratoPossuiRestricaoIngrediente(p, restricaoIngrediente) &&
        !pratoPossuiRestObg(diaCalendarioOrigem, p, utilizacaoPratoSemana) &&
        !pratoPossuiRestObg(diaCalendarioDestino, p, utilizacaoPratoSemana)
      );
    });
    return !!pratosPossiveis.length && !possuiGrupoDia;
  });
  if (diasCandidatos.length === 0) return false;
  let trocaRealizada = false;
  let segundaRodada = false;
  for (let index = 0; index < diasCandidatos.length; index++) {
    let dia = diasCandidatos[index];
    let pratosCandidatos = dia.pratos.filter(
      (p) =>
        p.prato &&
        groupSeqMap.get(pratoDia.gpprato) &&
        groupSeqMap.get(pratoDia.gpprato).includes(p.sequencial) /*&& p.sequencial == pratoDia.sequencial*/,
    );
    pratosCandidatos = pratosCandidatos.filter((p) => groupSeqMap.get(p.gpprato).includes(pratoDia.sequencial)); // pratos que podem ir para o sequencial q esta sendo trocado (origem)
    let diaAtual = getDay(dia.dia);
    let diaOrigem = getDay(pratoDia.data);
    let indexOrigemTabela = cardapioTabela[diaOrigem].findIndex(
      (p) => p.prato == pratoDia.prato && p.sequencial == pratoDia.sequencial,
    );
    const combOrigem = getComb(cardapioTabela[diaOrigem], pratoDia);
    let indexDiaSelecionado = cardapioDia.pratos.findIndex(
      (p) => p.prato && p.sequencial == pratoDia.sequencial && p.prato == pratoDia.prato,
    );
    let continueLoop = false;
    let breakLoop = false;
    trocaRealizada = pratosCandidatos.some((pratoCandidato) => {
      let seqDestino = pratoCandidato.sequencial;
      let indexCandidatoTabela = cardapioTabela[diaAtual].findIndex(
        (p) =>
          (p.prato == pratoCandidato.prato || p.codprato == pratoCandidato.prato) &&
          p.sequencial == pratoCandidato.sequencial,
      );
      let pratoCandidatoTabela = indexCandidatoTabela > -1 ? cardapioTabela[diaAtual][indexCandidatoTabela] : null;
      const combCandidato = getComb(cardapioTabela[diaAtual], pratoCandidato);
      const possuiCombinacao = !!combCandidato.length || !!combOrigem.length;
      if (pratoCandidatoTabela && !possuiCombinacao) {
        let cardapioDiaCandidato = cardapio.find((c) => c.dia == pratoCandidato.data);
        let indexDiaCandidato = cardapioDiaCandidato.pratos.findIndex(
          (p) => p.prato && p.sequencial == pratoCandidato.sequencial && p.prato == pratoCandidato.prato,
        );
        let possuiGrupoAntesDepois = false;
        let possuiMesmoGrupoNoDia = false;
        if (indexDiaCandidato > -1 && indexDiaSelecionado > -1 && indexOrigemTabela > -1 && indexCandidatoTabela > -1) {
          if (!segundaRodada) {
            if (diaAtual > 1 && diaAtual !== cardapioTabela.length - 1) {
              let possuiGrupoAntes = cardapioTabela[diaAtual - 1].some(
                (p) => p.gpprato == pratoDia.gpprato && isOtherDishOrDay(pratoDia, p),
              ); //dia anterior
              let possuiGrupoDepois = cardapioTabela[diaAtual + 1].some(
                (p) => p.gpprato == pratoDia.gpprato && isOtherDishOrDay(pratoDia, p),
              ); // dia posterior
              if (possuiGrupoAntes || possuiGrupoDepois) possuiGrupoAntesDepois = true;
            } else if (diaAtual == 1) {
              // primeiroDia
              let possuiGrupoDepois = cardapioTabela[diaAtual + 1].some(
                (p) => p.gpprato == pratoDia.gpprato && isOtherDishOrDay(pratoDia, p),
              );
              if (possuiGrupoDepois) possuiGrupoAntesDepois = true;
            } else if (diaAtual == cardapioTabela.length - 1) {
              // ultimoDia
              let possuiGrupoAntes = cardapioTabela[diaAtual - 1].some(
                (p) => p.gpprato == pratoDia.gpprato && isOtherDishOrDay(pratoDia, p),
              );
              if (possuiGrupoAntes) possuiGrupoAntesDepois = true;
            }
            if (!possuiGrupoAntesDepois) {
              if (diaOrigem > 1 && diaOrigem !== cardapioTabela.length - 1) {
                let possuiGrupoAntes = cardapioTabela[diaOrigem - 1].some(
                  (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
                ); //dia anterior
                let possuiGrupoDepois = cardapioTabela[diaOrigem + 1].some(
                  (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
                ); // dia posterior
                if (possuiGrupoAntes || possuiGrupoDepois) possuiGrupoAntesDepois = true;
              } else if (diaOrigem == 1) {
                // primeiroDia
                let possuiGrupoDepois = cardapioTabela[diaOrigem + 1].some(
                  (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
                ); // dia posterior
                if (possuiGrupoDepois) possuiGrupoAntesDepois = true;
              } else if (diaOrigem == cardapioTabela.length - 1) {
                // ultimoDia
                let possuiGrupoAntes = cardapioTabela[diaOrigem - 1].some(
                  (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
                ); //dia anterior
                if (possuiGrupoAntes) possuiGrupoAntesDepois = true;
              }
            }
          }
          // no dia que esta envolvido na troca de pratos ja existe prato com o msm grupo de prato a ser recebido.
          if (
            cardapioTabela[diaOrigem].some(
              (p) => p.gpprato == pratoCandidato.gpprato && isOtherDishOrDay(pratoCandidato, p),
            ) ||
            cardapioTabela[diaAtual].some((p) => p.gpprato == pratoDia.gpprato && isOtherDishOrDay(pratoDia, p))
          ) {
            possuiMesmoGrupoNoDia = true;
          }

          if (force) {
            if (possuiGrupoAntesDepois && index == diasCandidatos.length - 1 && !segundaRodada) {
              index = -1;
              segundaRodada = true;
              continueLoop = true;
              return true; //break loop some
            } else if (possuiGrupoAntesDepois && !segundaRodada) {
              continueLoop = true;
              return true; //break loop some
            }
          } else if (possuiGrupoAntesDepois) {
            continueLoop = true;
            return true; //break loop some
          }
          if (possuiMesmoGrupoNoDia) {
            if (force && index == diasCandidatos.length - 1 && !segundaRodada) {
              index = -1;
              segundaRodada = true;
            }
            continueLoop = true;
            return true; //break loop some
          }
          registraLog(
            log,
            pratoDia.data,
            dia.dia,
            pratoDia,
            pratoCandidato,
            `Segunda rodada: ${segundaRodada ? 'S' : 'N'}`,
          );
          registraLog(
            log,
            dia.dia,
            pratoDia.data,
            pratoCandidato,
            pratoDia,
            `Segunda rodada: ${segundaRodada ? 'S' : 'N'}`,
          );

          let aux = copyObj(cardapioDia.pratos[indexDiaSelecionado]);
          aux.data = cardapioDiaCandidato.dia;

          aux.sequencial = seqDestino;
          cardapioDiaCandidato.pratos[indexDiaCandidato].data = cardapioDia.dia;
          cardapioDiaCandidato.pratos[indexDiaCandidato].sequencial = seqOrigem;
          cardapioDia.pratos[indexDiaSelecionado] = cardapioDiaCandidato.pratos[indexDiaCandidato];
          cardapioDiaCandidato.pratos[indexDiaCandidato] = aux;

          aux = copyObj(cardapioTabela[diaOrigem][indexOrigemTabela]);
          aux.data = cardapioDiaCandidato.dia;
          aux.sequencial = seqDestino;
          cardapioTabela[diaAtual][indexCandidatoTabela].data = cardapioDia.dia;
          cardapioTabela[diaAtual][indexCandidatoTabela].sequencial = seqOrigem;
          cardapioTabela[diaOrigem][indexOrigemTabela] = cardapioTabela[diaAtual][indexCandidatoTabela];
          cardapioTabela[diaAtual][indexCandidatoTabela] = aux;
          breakLoop = true;

          return true; //break loop some
        } else {
          console.log('NAO_SUBSTITUIDO');
        }
      } else if (possuiCombinacao) {
        // Combinação atualmente eh feita a troca apenas dentro de um mesmo sequencial
        let diaCalendarioOrigem = originalData.calendario.data.find((d) => d.DTDATA.split(' ')[0] == dia.dia); // dia origem
        let diaCalendarioDestino = originalData.calendario.data.find((d) => d.DTDATA.split(' ')[0] == pratoDia.data); // dia destino
        let isValidCombCandidato = combCandidato.every((p) => {
          let pratoDaTroca = cardapioTabela[diaOrigem].find((prato) => prato.sequencial == p.sequencial);
          return (
            pratoDaTroca &&
            !cardapioTabela[diaOrigem].find(
              (prato) => prato.gpprato == p.gpprato && prato.sequencial != p.sequencial,
            ) &&
            !cardapioTabela[diaAtual].find(
              (prato) => prato.gpprato == pratoDaTroca.gpprato && prato.sequencial != pratoDaTroca.sequencial,
            ) &&
            !pratoPossuiRestricaoCombinacao(p, resPratoCombinacao) &&
            !pratoPossuiRestricaoIngrediente(p, restricaoIngrediente) &&
            !pratoPossuiRestObg(diaCalendarioOrigem, p, utilizacaoPratoSemana) &&
            !pratoPossuiRestObg(diaCalendarioDestino, p, utilizacaoPratoSemana) &&
            //validações relativas aos pratos q serao trocados pelas combinações acima testadas
            !pratoPossuiRestricaoCombinacao(pratoDaTroca, resPratoCombinacao) &&
            !pratoPossuiRestricaoIngrediente(pratoDaTroca, restricaoIngrediente) &&
            !pratoPossuiRestObg(diaCalendarioOrigem, pratoDaTroca, utilizacaoPratoSemana) &&
            !pratoPossuiRestObg(diaCalendarioDestino, pratoDaTroca, utilizacaoPratoSemana)
          );
        });
        let isValidCombOrigem = combOrigem.every((p) => {
          let pratoDaTroca = cardapioTabela[diaAtual].find((prato) => prato.sequencial == p.sequencial);
          return (
            pratoDaTroca &&
            !cardapioTabela[diaAtual].find((prato) => prato.gpprato == p.gpprato && prato.sequencial != p.sequencial) &&
            !cardapioTabela[diaOrigem].find(
              (prato) => prato.gpprato == pratoDaTroca.gpprato && prato.sequencial != pratoDaTroca.sequencial,
            ) &&
            !pratoPossuiRestricaoCombinacao(p, resPratoCombinacao) &&
            !pratoPossuiRestricaoIngrediente(p, restricaoIngrediente) &&
            !pratoPossuiRestObg(diaCalendarioOrigem, p, utilizacaoPratoSemana) &&
            !pratoPossuiRestObg(diaCalendarioDestino, p, utilizacaoPratoSemana) &&
            //validações relativas aos pratos q serao trocados pelas combinações acima testadas
            !pratoPossuiRestricaoCombinacao(pratoDaTroca, resPratoCombinacao) &&
            !pratoPossuiRestricaoIngrediente(pratoDaTroca, restricaoIngrediente) &&
            !pratoPossuiRestObg(diaCalendarioOrigem, pratoDaTroca, utilizacaoPratoSemana) &&
            !pratoPossuiRestObg(diaCalendarioDestino, pratoDaTroca, utilizacaoPratoSemana)
          );
        });
        if (isValidCombCandidato && isValidCombOrigem) {
          //? combinação tem que validar todos os pratos possíveis para cada prato combinante
          //? se cada prato combinante puder ser trocado por um outro prato, então a combinação é válida
          if (!segundaRodada) {
            let grupoDiaConsecutivo = combOrigem.some((p) => {
              let pratoDaTroca = cardapioTabela[diaAtual].find((prato) => prato.sequencial == p.sequencial);
              return verificaGrupoDiasConsecutivos(cardapioTabela, p, pratoDaTroca, diaOrigem, diaAtual);
            });
            if (!grupoDiaConsecutivo) {
              grupoDiaConsecutivo = combCandidato.some((p) => {
                let pratoDaTroca = cardapioTabela[diaOrigem].find((prato) => prato.sequencial == p.sequencial);
                return verificaGrupoDiasConsecutivos(cardapioTabela, pratoDaTroca, p, diaAtual, diaOrigem);
              });
            }
            if (force) {
              if (grupoDiaConsecutivo && index == diasCandidatos.length - 1 && !segundaRodada) {
                index = -1;
                segundaRodada = true;
                continueLoop = true;
                return true; //break loop some
              } else if (grupoDiaConsecutivo && !segundaRodada) {
                continueLoop = true;
                return true; //break loop some
              }
            } else if (grupoDiaConsecutivo) {
              continueLoop = true;
              return true; //break loop some
            }
          }
        }
      }
      return trocaRealizada;
    });
    if (continueLoop) {
      trocaRealizada = false;
      continue;
    } else if (breakLoop) {
      trocaRealizada = true;
      break;
    } else {
      trocaRealizada = false;
    }
  }
  return trocaRealizada;
};
this.dishCost = (vlrPrato, percCons, comensais) => {
  return (parseFloat(vlrPrato) || 0) * ((parseFloat(percCons) || 0) / 100) * parseFloat(comensais || 0);
};
const recalculaPu = (perfectCard) => {
  let comensalAcumulado = 0;
  let custoAcumulado = 0;
  perfectCard.cardapio.datas.forEach((dia) => {
    let mediaCusto = 0;
    let puDia = dia.pratos.reduce((acc, cv) => {
      acc += this.dishCost(cv.VLR_CUSTO_PRATO, cv.percentualConsumo, dia.comensais) / parseInt(dia.comenDIV);
      mediaCusto += (parseFloat(cv.VLR_CUSTO_PRATO) || 0) * ((parseFloat(cv.percentualConsumo) || 0) / 100);
      return acc;
    }, 0);
    dia.mediaCusto = mediaCusto;
    custoAcumulado += (puDia || 0) * (dia.comensais ? parseInt(dia.comensais) : 0);
    comensalAcumulado += parseInt(dia.comensais || 0);
    let puAcum = custoAcumulado / comensalAcumulado;
    perfectCard.tabela.pudia_puacu[getDay(dia.dia)] = [puDia, puAcum];
  });
};

this.execute = (perfectCard, originalData) => {
  let calendario = originalData.calendario.data;
  DataManagement.setDatePropsOnCalendario(calendario);
  this.seqMap = perfectCard.tabela.pratos[0].reduce((map, coluna) => {
    map.set(coluna.seq, coluna.name);
    return map;
  }, new Map());
  let analiseCardapio = {};
  let cardapio = perfectCard.cardapio;
  let seqsMap = new Map();
  let groupSeqMap = new Map();
  originalData.configuracao_servico.data.forEach(({ NUM_SEQUENCIA, NUM_GRP_PRATO }) => {
    if (!seqsMap.has(NUM_SEQUENCIA)) seqsMap.set(NUM_SEQUENCIA, 1);
    else seqsMap.set(NUM_SEQUENCIA, seqsMap.get(NUM_SEQUENCIA) + 1);
    if (!groupSeqMap.has(NUM_GRP_PRATO)) groupSeqMap.set(NUM_GRP_PRATO, [NUM_SEQUENCIA.toString()]);
    else groupSeqMap.set(NUM_GRP_PRATO, groupSeqMap.get(NUM_GRP_PRATO).concat(NUM_SEQUENCIA.toString()));
  });
  let { groupMap, gruposDuplicadosDia } = analisaCardapio(cardapio, seqsMap);

  //grupos repetidos em dias consecutivos
  for ([group, diasAgrup] of groupMap) {
    diasAgrup.forEach((dias) => {
      let diasOrdenados = priorizaDiasIntercalados(dias);
      let diasOrdenadosBkp = [...diasOrdenados];
      let i = 0;
      do {
        let dia = diasOrdenados[i];
        let { cardapio: cardapioSemana, indexCardDia } = retornaCardapioSemana(cardapio, dia, diasOrdenados);
        if (cardapioSemana.length && indexCardDia > -1) {
          let pratosDia = cardapio.datas[indexCardDia].pratos.filter((p) => p.gpprato == group); // pratos que vao tentar ser trocados
          let cardapioTabela = perfectCard.tabela.pratos;
          let trocaRealizada = false;
          pratosDia.forEach((pratoDia) => {
            trocaRealizada =
              realizaTrocaPrato(
                cardapio.datas,
                cardapio.datas[indexCardDia],
                cardapioTabela,
                cardapioSemana,
                pratoDia,
                originalData,
                groupSeqMap,
                analiseCardapio,
              ) || trocaRealizada;
          });
          if (trocaRealizada) {
            diasOrdenadosBkp.splice(
              diasOrdenadosBkp.findIndex((diaAtual) => diaAtual == dia),
              1,
            );
          }
        }
        i++;
      } while (i < diasOrdenados.length && possuiDiasConsecutivos(diasOrdenadosBkp));
    });
  }

  //grupos repetidos em mesmo dia
  Array.from(gruposDuplicadosDia.keys()).forEach((dia) => {
    let group = gruposDuplicadosDia.get(dia);
    let { cardapio: cardapioSemana, indexCardDia } = retornaCardapioSemana(cardapio, dia, [], true);
    if (cardapioSemana.length && indexCardDia > -1) {
      let cardapioDia = cardapio.datas[indexCardDia];
      let pratos = cardapioDia.pratos.filter((p) => p.gpprato == group); // pratos para tentar retirar do cardapio do dia
      for (let i = 0; i < pratos.length; i++) {
        let pratoDia = pratos[i];
        let cardapioTabela = perfectCard.tabela.pratos;
        let trocaRealizada = realizaTrocaPrato(
          cardapio.datas,
          cardapioDia,
          cardapioTabela,
          cardapioSemana,
          pratoDia,
          originalData,
          groupSeqMap,
          analiseCardapio,
          true,
        );
        if (trocaRealizada) break;
      }
    }
  });
  recalculaPu(perfectCard);
  perfectCard.tabela.log.analiseCardapio = analiseCardapio;
};
