const moment = require('moment');
const IngredientValidation = require('./IngredientValidation');

this.dishCost = (vlrPrato, percCons, comensais) => {
    return (parseFloat(vlrPrato) || 0) * ((parseFloat(percCons) || 0)/100) * parseFloat(comensais || 0);
};

this.getCombinationRestrictionMap = ( combination, restriction ) => {
    const combRestMap = new Map();
    combRestMap.set('COMB', {DISH: new Set(), GROUP: new Set()});
    combRestMap.set('REST', {DISH: new Set(), GROUP: new Set()});
    combination.forEach( comb => {
        combRestMap.get('COMB').DISH.add(comb.COD_PRATO);
        combRestMap.get('COMB').DISH.add(comb.COD_PRATO_COMB);

        combRestMap.get('COMB').GROUP.add(comb.NRGRUPOPRATOSERV);
        combRestMap.get('COMB').GROUP.add(comb.NRGRUPOPRATOSERV_COMB);
    });
    restriction.forEach( rest => {
        combRestMap.get('REST').DISH.add(rest.COD_PRATO);
        combRestMap.get('REST').GROUP.add(rest.NRGRUPOPRATOSERV);
    });
    return combRestMap;
};

this.isCombOrRest = ( dish, combRestMap ) => {
    const {GROUP: GROUP_COMB, DISH: DISH_COMB} = combRestMap.get('COMB');
    const {GROUP: GROUP_REST, DISH: DISH_REST} = combRestMap.get('REST');
    let hasGroupCombRest = false;
    let hasDishCombRest = false;
    if(dish.gpprato) {
        hasGroupCombRest = GROUP_COMB.has(dish.gpprato) || GROUP_REST.has(dish.gpprato);
    }
    hasDishCombRest = DISH_COMB.has(dish.prato) || DISH_REST.has(dish.prato);
    return hasDishCombRest || hasGroupCombRest;
};

this.hasDishAndCost = ( dish ) => {
    return !!(dish.prato && dish.VLR_CUSTO_PRATO);
};

this.hasOptionsToChange = ( dish ) => {
    return !!(Array.isArray(dish.pratosBKP) && dish.pratosBKP.length);
};

this.isValidFreqDish = ( dish, frequency ) => {
    const freq = frequency.find( freq => freq.COD_PRATO === dish.prato);
    return freq && freq.cur && freq.MIN_MES_PRATO ? freq.cur > freq.MIN_MES_PRATO : true;
};

this.isValidFreqCandidate = ( originalDish, dish, frequency ) => {
    const freq = frequency.find( freq => freq.NUM_GRP_PRATO === dish.gpprato );
    const freqDish = frequency.find( freq => freq.COD_PRATO === dish.prato );
    const weekNumber = moment(dish.data, 'DD/MM/YYYY').week();
    const groupInWeek = freq && freq.curWeekUsage ? freq.curWeekUsage[weekNumber] : 0;
    const dishInWeek =  freqDish && freqDish.curWeekUsage ? freqDish.curWeekUsage[weekNumber] : 0;
    const isValidFreqMonthWeekGroup = (freq, groupInWeek ) => {
        if( !freq || !freq.MAX_MES_GRP || !freq.cur) return true;
        return ( freq.cur < freq.MAX_MES_GRP )
            && (!groupInWeek || (groupInWeek < freq.MAX_SEM_GRP))
    };
    const isValidFreqMonthWeekDish = (freq, dishInWeek) => {
        if(!freq || !freq.MAX_MES_PRATO || !freq.cur) return true;
        return (freq.cur < freq.MAX_MES_PRATO)
            && (!dishInWeek || (dishInWeek < freq.MAX_SEMANA))
    };
    return (originalDish.gpprato == dish.gpprato || isValidFreqMonthWeekGroup(freq, groupInWeek))
        && isValidFreqMonthWeekDish(freqDish, dishInWeek)
};

this.getDishHighCost = ( dishCollection, combRestMap, freq, dishToIgnoreChange = new Map() ) => {
    let dishHighCost = {VLR_CUSTO_PRATO: 0};
    let indexDish = -1;
    dishCollection.forEach( (dish, index) => {
        if( !dishToIgnoreChange.has(dish.data+dish.prato)
            && this.hasDishAndCost(dish)
            && this.hasOptionsToChange( dish ) && parseFloat(dish.VLR_CUSTO_PRATO) > parseFloat(dishHighCost.VLR_CUSTO_PRATO)
            && !this.isCombOrRest( dish, combRestMap )
            && this.isValidFreqDish( dish, freq )
        ){
            dishHighCost = dish;
            indexDish = index;
        }
    });
    return indexDish > -1 ? {dish: dishHighCost, index: indexDish} : null;
};

this.isValidCombCandidate = ( menuToday, candidate, combination ) => {
    let hasFilledComb = false;
    let combinations = combination.filter( comb => comb.COD_PRATO == candidate.prato || comb.NRGRUPOPRATOSERV == candidate.gpprato);
    if( combinations.length ) {
        hasFilledComb = false && combinations.some( comb => {
            let [ COLUMN_1, COLUMN_2 ] = typeof comb.SEQ == 'string' ? comb.SEQ.match(/\d+/gi) || [] : comb.SEQ;
            if(COLUMN_1 && COLUMN_2) {
                return (COLUMN_1 != candidate.sequencial) || menuToday.some( dish => {
                    dish.sequencial == COLUMN_2 && dish.prato == dishComb.COD_PRATO;
                });
            }
            return false;
        }); // revisar o que seria "dishComb" e verificar se a função 'some' eh apropriada com condicional COLUMN_1 != seq
    }
    return ( combinations.length === 0 ) || hasFilledComb;
};

this.isValidRestCandidate = ( menuToday, candidate, restriction ) => {
    let hasRestriction = null;
    let restrictions = restriction.filter( rest => {
        return ( rest.CDTIPOPRAT && rest.CDTIPOPRAT == candidate.COD_TIP_PRATO)
                || (rest.COD_PRATO && rest.COD_PRATO == candidate.prato)
                || (rest.NRGRUPOPRATOSERV && rest.NRGRUPOPRATOSERV == candidate.gpprato);
    });
    let restrictions_comb = restriction.filter( rest => {
        return ( rest.CDTIPOPRAT && rest.CDTIPOPRAT == candidate.COD_TIP_PRATO)
                || (rest.COD_PRATO_COMB && rest.COD_PRATO_COMB == candidate.prato)
                || (rest.NRGRUPOPRATOSERV_COMB && rest.NRGRUPOPRATOSERV_COMB == candidate.gpprato);
    });
    if( restrictions.length ) {
        hasRestriction = restrictions.some( rest => {
            let [ COLUMN_1, COLUMN_2 ] = rest.SEQ.match(/\d+/gi) || [];
            if(COLUMN_1 && COLUMN_2) {
                return (COLUMN_1 == candidate.sequencial) && menuToday.some( dish => {
                    return dish.sequencial == COLUMN_2 &&
                    (
                        (rest.COD_GRP_TIP_PRATO_COMB == dish.COD_GRP_TIP_PRATO && !rest.NRGRUPOPRATOSERV_COMB && !rest.COD_PRATO_COMB) 
                        || rest.NRGRUPOPRATOSERV_COMB == dish.gpprato
                        || rest.COD_PRATO_COMB == dish.prato
                    );
                });
            }
        });
    } else {
        return true;
    }

    if( restrictions_comb.length ) {
        hasRestriction = restrictions_comb.some( rest => {
            let [ COLUMN_1, COLUMN_2 ] = rest.SEQ.match(/\d+/gi) || [];
            if(COLUMN_1 && COLUMN_2) {
                return (COLUMN_2 == candidate.sequencial) && menuToday.some( dish => {
                    return dish.sequencial == COLUMN_1 && 
                    (
                        (rest.COD_GRP_TIP_PRATO == dish.COD_GRP_TIP_PRATO && !rest.NRGRUPOPRATOSERV && !rest.COD_PRATO) 
                        || rest.NRGRUPOPRATOSERV == dish.gpprato 
                        || rest.COD_PRATO == dish.prato
                    );
                });
            }
        });
        return !hasRestriction;
    }
    return true;
};

this.getDishLowCost = ( CDFILIAL, menu, dishCollection, currentDish, freq, dishFreqCurWeek, restIng, combRest, onlySameGroup, dishesToIgnore ) => {
    const { today } = menu;
    const maxValue = currentDish.VLR_CUSTO_PRATO;
    const MAX_DISH_PER_WEEK = 2;
    const onlyRequiredDish = currentDish.utilizacao === 'O';
    let selectedDish = {VLR_CUSTO_PRATO: maxValue};
    dishCollection.forEach( dish => {
        let isLowestCost = parseFloat(dish.VLR_CUSTO_PRATO) < parseFloat(selectedDish.VLR_CUSTO_PRATO);
        if(isLowestCost
            && ((dishFreqCurWeek[getDishCode(dish, CDFILIAL)] || 0) < MAX_DISH_PER_WEEK )
            && (onlyRequiredDish ? dish.utilizacao === 'O' : true)
            && (onlySameGroup ? dish.gpprato == currentDish.gpprato : true)
            && (!dishesToIgnore.has(getDishCode(dish, CDFILIAL)))
            && this.isValidFreqCandidate(currentDish, dish, freq)
            && this.isValidCombCandidate(today, dish, combRest.COMB)
            && this.isValidRestCandidate(today, dish, combRest.REST)
            && IngredientValidation.isAllowedIngredient(menu, dish.INGR_PRINCIPAL_1, dish.INGR_PRINCIPAL_2, dish.COD_TIP_PRATO, restIng)) {
            selectedDish = dish;
        }
    });
    return selectedDish.VLR_CUSTO_PRATO === maxValue ? null : selectedDish;
};

this.copyObj = obj => JSON.parse(JSON.stringify(obj));

this.getFilteredDishOptions = (dishCollection, today, menuYesterdayAndTomorrow, hasPortion) => {
    if(hasPortion){
        const hasGroup = ( dishCollection, groupId ) => dishCollection.some( dish => dish.gpprato === groupId);
        return dishCollection.filter( dish => dish.prato && !hasGroup(menuYesterdayAndTomorrow, dish.gpprato) && !today.find(tDish => tDish.gpprato == dish.gpprato))
    } else {
        const hasDish = ( dishCollection, dishId ) => dishCollection.some(dish => dish.prato === dishId);
        return dishCollection.filter( dish => dish.prato && !hasDish(menuYesterdayAndTomorrow, dish.prato) && !today.find(tDish => tDish.gpprato == dish.gpprato));
    }
};

function getDishCode( dish, CDFILIAL ){
    if(dish.prato) {
        let dishCode = dish.prato[0] === 'P' ? dish.prato : null;
        if(dishCode == null) {
            if(dish.prato.indexOf(CDFILIAL) !== 0){
                dishCode = CDFILIAL+dish.prato;
            }
        }
        return dishCode;
    } else {
        return null;
    }
}

function isEqualDish(dishCode1, dishCode2, CDFILIAL){
    if(typeof dishCode1 !== 'string' || typeof dishCode2 !== 'string' || typeof CDFILIAL !== 'string') return false;
    let dishes = ['', ''];
    if(dishCode1 == dishCode2) return true;
    dishes[0] = dishCode1[0] == 'P' ? dishCode1.substring(1) : dishCode1;
    dishes[0] = dishCode1 == dishes[0] && dishCode1.indexOf(CDFILIAL) === 0 ? dishCode1.substring(CDFILIAL.length) : dishes[0];
    dishes[1] = dishCode2[0] == 'P' ? dishCode2.substring(1) : dishCode2;
    dishes[1] = dishCode2 == dishes[1] && dishCode2.indexOf(CDFILIAL) === 0 ? dishCode2.substring(CDFILIAL.length) : dishes[1];
    return dishes[0] === dishes[1];
}

this.changeDish = ( CDFILIAL, menuDate, menu, puDia, combRest, combRestMap, portionMap, freq, dishFreqPerWeek,restIng, comensaisMap, dishToIgnoreChange = new Map() ) => {
    const { yesterday, today, tomorrow } = menu;
    dishesToIgnore = new Set();
    yesterday.concat(tomorrow).forEach( dish => {
        let dishCode = getDishCode(dish, CDFILIAL);
        if(dishCode) dishesToIgnore.add(dishCode);
    });
    let { dish, index } = this.getDishHighCost( today, combRestMap, freq, dishToIgnoreChange ) || {VLR_CUSTO_PRATO: 0};
    let curFreq = dish && dish.gpprato ? freq.find( fr => fr.NUM_GRP_PRATO == dish.gpprato) : null;
    const {VLR_CUSTO_PRATO: CURRENT_COST} = dish || {VLR_CUSTO_PRATO: null};
    if( CURRENT_COST ) {
        dishToIgnoreChange.set(dish.data+dish.prato, '');
        const onlySameGroup = curFreq && curFreq.cur <= curFreq.MIN_MES_GRP;
        const hasPortion = portionMap.has(dish.COD_GRP_TIP_PRATO);
        const filteredOptions = this.getFilteredDishOptions(dish.pratosBKP, today, yesterday.concat(tomorrow),hasPortion);
        const weekNumber = moment(menuDate, 'DD/MM/YYYY').week();
        let selectedDish = this.getDishLowCost(CDFILIAL, menu, filteredOptions, dish, freq, dishFreqPerWeek[weekNumber], restIng, combRest, onlySameGroup, dishesToIgnore);
        if( selectedDish ) {
            let infoChange = {
                DAY: menuDate,
                SEQ: selectedDish.sequencial, 
                OLD_DISH: dish.prato,
                NEW_DISH: selectedDish.prato
            };
            let {pratosBKP, ...dishProps} = dish;
            let newPratosBKP = pratosBKP.filter( dish => dish.prato !== selectedDish.prato);
            newPratosBKP.push(dishProps);
            let newDish = this.copyObj(selectedDish);
            newDish.pratosBKP = newPratosBKP;
            newDish.otherOptions = dish.otherOptions;
            if( !newDish.codprato ) {
                let dishCode = newDish.prato;
                if(dishCode[0] == 'P') {
                    dishCode = newDish.prato;
                    newDish.prato = newDish.prato.substring(1);
                } else if(newDish.unidadePrato == 'P'){
                    dishCode = 'P' + dishCode;
                }
                newDish.codprato = dishCode;
            }
            today[index] = newDish;
            let newPuDia = today.reduce((acc, cv)=>{
                acc += (this.dishCost(cv.VLR_CUSTO_PRATO, cv.percentualConsumo, comensaisMap.comensais))/comensaisMap.comenDIV;
                return acc; 
            }, 0);
            if(selectedDish.gpprato !== dish.gpprato) {
                let curFreqCandidate = selectedDish && selectedDish.gpprato ? freq.find( fr => fr.NUM_GRP_PRATO == selectedDish.gpprato) : null;
                if(curFreqCandidate) {
                     curFreqCandidate.cur++;
                }
            }
            return { newPuDia, infoChange };
        } else {
            return { newPuDia: puDia };
        }
    } 
    return { newPuDia: puDia };
}

this.getPortionMap = ( portionDataSet ) => {
    const portionMap = new Map();
    portionDataSet.forEach( ( { CDGRTIPOPR, VRCONJUNTOPORCAO } ) => {
        portionMap.set(CDGRTIPOPR, VRCONJUNTOPORCAO);
    });
    return portionMap;
}

this.analizeGeneratedCard = ( generatedCard, originalData, logChanges = {}, dishToIgnoreChange = new Map() ) => {
    let monthYearMenu = generatedCard.cardapio.mes.substring(2);
    let CDFILIAL = generatedCard.cardapio.unidade;
    let puDiaMap = generatedCard.tabela.pudia_puacu.map( ([puDia], index)=> ({dia: index, puDia})).sort((a,b) => {
        return a.puDia < b.puDia || typeof a.puDia === 'string' ? 1 : -1;
    });
    const lastWeekProds = originalData.data.lastWeekProds.data;
    let combRestMap = this.getCombinationRestrictionMap(originalData.data.prato_combinacao.data, originalData.data.res_prato_combinacao.data);
    let combRest = {
        REST: originalData.data.res_prato_combinacao.data, 
        COMB: originalData.data.prato_combinacao.data
    };
    let restIng = originalData.data.restricao_de_repeticao_ingrediente.data;
    let portionMap = this.getPortionMap(originalData.data.conjunto_porcao.data);
    
    puDiaMap.pop(); // remove header item;
    let {pu_aculm, comensal_total} = generatedCard;
    let newPuAc = pu_aculm;
    let { meta } = originalData.meta;
    let changes = 0;
    let index = 0;
    let daysInMonth = generatedCard.tabela.pratos.length - 1;
    let freq = generatedCard.frequencias;
    let { comensais, dishFreqInMonth, dishFreqPerWeek } = generatedCard.cardapio.datas.reduce( (acc, diaCard) => {
        let diaMoment = moment(diaCard.dia, 'DD/MM/YYYY');
        let dia = parseInt(diaMoment.format('DD'));
        let weekNumber = diaMoment.week();
        acc.comensais[dia] = {comenDIV: diaCard.comenDIV, comensais: diaCard.comensais};
        acc.dishFreqPerWeek[weekNumber] = acc.dishFreqPerWeek[weekNumber] || {};
        diaCard.pratos.forEach( dish => {
            acc.dishFreqInMonth[dish.prato] = acc.dishFreqInMonth[dish.prato] ? acc.dishFreqInMonth[dish.prato]+1 : 1;
            acc.dishFreqPerWeek[weekNumber][getDishCode(dish, CDFILIAL)] = acc.dishFreqPerWeek[weekNumber][getDishCode(dish, CDFILIAL)] ? acc.dishFreqPerWeek[weekNumber][getDishCode(dish, CDFILIAL)]+1 : 1;
        });
        
        return acc;
    }, {comensais: {}, dishFreqInMonth: {}, dishFreqPerWeek: {}});
    lastWeekProds.forEach( dish => {
        let weekNumber = moment(dish.DTDATA, 'DD/MM/YYYY').week();
        dishFreqPerWeek[weekNumber][getDishCode({prato: dish.COD_PRATO}, dish.COD_FILIAL)] = dishFreqPerWeek[weekNumber][getDishCode({prato: dish.COD_PRATO}, dish.COD_FILIAL)] ? dishFreqPerWeek[weekNumber][getDishCode({prato: dish.COD_PRATO}, dish.COD_FILIAL)]+1 : 1;
    });
    while(newPuAc-meta > 0 && index <= daysInMonth){
        let { dia, puDia } = puDiaMap[index] || {dia: null, puDia: null};
        if(!isNaN(puDia) && dia !== null && puDia !== null && puDia > 0) {
            let menu = generatedCard.tabela.pratos;
            let menuDate = dia.toString().padStart(2, '0') + monthYearMenu;
            let yesterday = dia == 1 ? [] : menu[dia-1];
            let tomorrow = dia == daysInMonth ? [] : menu[dia+1];
            const menuRange = { yesterday, today: menu[dia], tomorrow };
            let {newPuDia, infoChange = null} = this.changeDish(generatedCard.cardapio.unidade, menuDate, menuRange, puDia, combRest, combRestMap, portionMap, freq, dishFreqPerWeek, restIng, comensais[dia], dishToIgnoreChange);
            let custoAcumulado = 0;
            let comensaisAcumulados = 0;
            generatedCard.tabela.pudia_puacu[dia][0] = newPuDia;
            if( infoChange ) {// houve troca de prato
                const {
                    DAY,
                    SEQ,
                    NEW_DISH,
                    OLD_DISH
                } = infoChange;
                const weekNumber = moment(DAY, 'DD/MM/YYYY').week();
                dishFreqInMonth[OLD_DISH]--;
                dishFreqInMonth[NEW_DISH] = dishFreqInMonth[NEW_DISH] ? dishFreqInMonth[NEW_DISH]+1 : 1;
                dishFreqPerWeek[weekNumber][getDishCode({prato: OLD_DISH}, CDFILIAL)]--;
                dishFreqPerWeek[weekNumber][getDishCode({prato: NEW_DISH}, CDFILIAL)] = dishFreqPerWeek[weekNumber][getDishCode({prato: NEW_DISH}, CDFILIAL)] ? dishFreqPerWeek[weekNumber][getDishCode({prato: NEW_DISH}, CDFILIAL)]+1 : 1;
                let menuDay = generatedCard.cardapio.datas.find( currentDay => currentDay.dia === DAY);
                if( menuDay ) {
                    let log = {};
                    let dishes = menuDay.pratos;
                    let indexOldDish = dishes.findIndex( dish => dish.sequencial == SEQ && isEqualDish(dish.prato, OLD_DISH, CDFILIAL));
                    let indexNewDish = dishes[indexOldDish].pratosBKP.findIndex( dish => dish.sequencial == SEQ && isEqualDish(dish.prato, NEW_DISH, CDFILIAL));
                    let {pratosBKP, otherOptions, data, ...oldDish} = dishes[indexOldDish];
                    let newDish = this.copyObj(pratosBKP[indexNewDish]);
                    log['Antes'] = dishes[indexOldDish].prato + ' ' + dishes[indexOldDish].nomePrato; 
                    log['Depois'] = newDish.prato + ' ' + newDish.nomePrato;
                    let { name } = generatedCard.tabela.pratos[0].find( typeDish => typeDish.seq == SEQ) || {name: ' - '};
                    log['Coluna'] = name;
                    newDish.otherOptions = otherOptions;
                    newDish.oc = dishFreqInMonth[NEW_DISH];
                    newDish.data = data;
                    let newPratosBKP = pratosBKP.filter(dish => dish.prato !== NEW_DISH);
                    newDish.pratosBKP = [oldDish].concat(this.copyObj(newPratosBKP));
                    dishes[indexOldDish] = newDish;
                    let {ING_USED, MDPREP_USED, TPCORT_USED} = dishes.reduce( (acc, {MODO_PREPARO, TP_CORTE}) => {
                        if( MODO_PREPARO ) acc.MDPREP_USED.add(MODO_PREPARO);
                        if( TP_CORTE ) acc.TPCORT_USED.add(TP_CORTE);
                        return acc;
                    }, {ING_USED: [], MDPREP_USED: new Set(), TPCORT_USED: new Set()});
                    menuDay.MDPREP_USED = Array.from(MDPREP_USED);
                    menuDay.TPCORT_USED = Array.from(TPCORT_USED);
                    menuDay.ING_USED = Array.from(ING_USED);
                    if(!logChanges.hasOwnProperty(DAY)) logChanges[DAY] = [];
                    logChanges[DAY].push(log);
                    changes++;
                }
            }
            generatedCard.tabela.pudia_puacu.forEach( (pu, dia) => {
                let [ puDia ] = pu;
                if(dia) {
                    let comensaisDia = comensais[dia];
                    let comensaisEstimados = comensaisDia ? parseInt(comensaisDia.comensais) : 0;
                    comensaisAcumulados += comensaisEstimados || 0;
                    custoAcumulado += (puDia || 0) * (comensaisEstimados || 0);
                    pu[1] = custoAcumulado/comensaisAcumulados;
                }
            });
            let lastPu = generatedCard.tabela.pudia_puacu.length - 1;
            newPuAc = generatedCard.tabela.pudia_puacu[lastPu][1];
        }
        index++;
    }
    newPuAc = parseFloat((typeof newPuAc == 'string' ? parseFloat(newPuAc) : newPuAc).toFixed(5));
    generatedCard.pu_aculm = parseFloat((typeof generatedCard.pu_aculm == 'string' ? parseFloat(generatedCard.pu_aculm) : generatedCard.pu_aculm).toFixed(5));
    const newTotalCost = parseInt(comensal_total) * newPuAc;
    generatedCard.pu_aculm = newPuAc;

    generatedCard.dif_custo = generatedCard.P_VRCUSTPADR - newPuAc;
    generatedCard.cust_total = newTotalCost;
    return changes;
}